#if 0
ctime -begin ${0%.*}.ctm
gcc -g -Wall -fsanitize=address -std=c99 $0 -o ${0%.*} hmml.a
ctime -end ${0%.*}.ctm
exit
#endif

#define DEBUG 1

// TODO(matt): Fully investigate the ClaimedMemory situation

typedef unsigned int bool;

#define TRUE 1
#define FALSE 0

#include <stdarg.h> // NOTE(matt): varargs
#include <stdio.h>  // NOTE(matt): printf, sprintf, vsprintf, fprintf, perror
#include <stdlib.h> // NOTE(matt): calloc, malloc, free
#include "hmmlib.h"

typedef struct
{
    char *Location;
    char *Ptr;
    char *ID;
    int  Size;
} buffer;

// TODO(matt): Consider putting the ref_info and quote_info into linked lists on the heap, just to avoid all the hardcoded sizes

typedef struct
{
    char Date[32];
    char Text[512];
} quote_info;

typedef struct
{
    char Timecode[8];
    int Identifier;
} identifier;

#define REF_MAX_IDENTIFIER 32

typedef struct
{
    char RefTitle[620];
    char ID[512];
    char URL[512];
    char Source[256];
    identifier Identifier[REF_MAX_IDENTIFIER];
    int IdentifierCount;
} ref_info;

typedef struct
{
    char Category[32];
    bool IsMedium;
} category_info;

#define ArrayCount(A) sizeof(A)/sizeof(*(A))

void
ClaimBuffer(char *MemoryArena, int *ClaimedMemory, buffer *Buffer, char *ID, int Size)
{
    Buffer->Location = MemoryArena + *ClaimedMemory;
    Buffer->Size = Size;
    Buffer->ID = ID;
    *ClaimedMemory += Buffer->Size;
    *Buffer->Location = '\0';
    Buffer->Ptr = Buffer->Location;
#if DEBUG
    printf("  Claimed: %s: %d\n"
            "    Total ClaimedMemory: %d\n\n", Buffer->ID, Buffer->Size, *ClaimedMemory);
#endif
}

void
DeclaimBuffer(buffer *Buffer, int *ClaimedMemory)
{
    *Buffer->Location = '\0';
    *ClaimedMemory -= Buffer->Size;
#if DEBUG
    printf("Declaimed: %s\n"
            "    Total ClaimedMemory: %d\n\n", Buffer->ID, *ClaimedMemory);
#endif
}

int
TimecodeToSeconds(char *Timecode)
{
    int HMS[3] = { 0, 0, 0 }; // 0 == Seconds; 1 == Minutes; 2 == Hours
    int Colons = 0;
    while(*Timecode)
    {
        //if((*Timecode < '0' || *Timecode > '9') && *Timecode != ':') { return FALSE; }
        
        if(*Timecode == ':')
        {
            ++Colons;
            //if(Colons > 2) { return FALSE; }
            for(int i = 0; i < Colons; ++i)
            {
                HMS[Colons - i] = HMS[Colons - (i + 1)];
            }
            HMS[0] = 0;
        }
        else
        {
            HMS[0] = HMS[0] * 10 + *Timecode - '0';
        }
        
        ++Timecode;
    }
    
    //if(HMS[0] > 59 || HMS[1] > 59 || Timecode[-1] == ':') { return FALSE; }
    
    return HMS[2] * 60 * 60 + HMS[1] * 60 + HMS[0];
}

void
CopyBuffer(buffer *Dest, buffer *Src)
{
    Src->Ptr = Src->Location;
    while(*Src->Ptr)
    {
        // TODO(matt)
        {
            if(Dest->Ptr - Dest->Location >= Dest->Size)
            {
                printf("Too big! Too big!\n");
                __asm__("int3");

            }
        }
        *Dest->Ptr++ = *Src->Ptr++;
    }
}

__attribute__ ((format (printf, 2, 3)))
void
CopyString(char Dest[], char *Format, ...)
{
    va_list Args;
    va_start(Args, Format);
    vsprintf(Dest, Format, Args);
    va_end(Args);
}

__attribute__ ((format (printf, 2, 3)))
void
CopyStringToBuffer(buffer *Dest, char *Format, ...)
{
    // TODO(matt):
    {
        if(Dest->Ptr - Dest->Location >= Dest->Size)
        {
            printf("Too big! Too big!\n");
            __asm__("int3");
        }
    }
    va_list Args;
    va_start(Args, Format);
    int Length = vsprintf(Dest->Ptr, Format, Args);
    va_end(Args);
    // TODO(matt):
    {
        if(Length + (Dest->Ptr - Dest->Location) >= Dest->Size)
        {
            printf("Too big! Too big!\n");
            __asm__("int3");
        }
    }
    Dest->Ptr += Length;
}

int
StringsDiffer(char *A, char *B)
{
    while(*A && *B && *A == *B)
    {
        ++A, ++B;
    }
    return *A - *B;
}

bool
StringsDifferL(char *A, char *B, int LengthofA)
{
    int i = 0;
    while(i < LengthofA && A[i] && A[i] == B[i])
    {
        ++i;
    }
    if(!A[i] && LengthofA == i)
    {
        return FALSE;
    }
    else
    {
        return TRUE;
    }
}

typedef struct
{
    unsigned int Hue:16;
    unsigned int Saturation:8;
    unsigned int Lightness:8;
} hsl_colour;

hsl_colour
CharToColour(char Char)
{
    hsl_colour Colour;

    if(Char >= 'a' && Char <= 'z')
    {
        Colour.Hue = (((float)Char - 'a') / ('z' - 'a') * 360);
        Colour.Saturation = (((float)Char - 'a') / ('z' - 'a') * 26 + 74);
    }
    else if(Char >= 'A' && Char <= 'Z')
    {
        Colour.Hue = (((float)Char - 'A') / ('Z' - 'A') * 360);
        Colour.Saturation = (((float)Char - 'A') / ('Z' - 'A') * 26 + 74);
    }
    else if(Char >= '0' && Char <= '9')
    {
        Colour.Hue = (((float)Char - '0') / ('9' - '0') * 360);
        Colour.Saturation = (((float)Char - '0') / ('9' - '0') * 26 + 74);
    }
    else
    {
        Colour.Hue = 180;
        Colour.Saturation = 50;
    }

    return Colour;
}

char *
StringToColourHash(buffer *Buffer, char *String)
{
    hsl_colour Colour = {0, 0, 26};

    int i;
    for(i = 0; String[i]; ++i)
    {
        Colour.Hue += CharToColour(String[i]).Hue;
        Colour.Saturation += CharToColour(String[i]).Saturation;
    }

    Colour.Hue = Colour.Hue % 360;
    Colour.Saturation = Colour.Saturation % 26 + 74;
    Buffer->Ptr = Buffer->Location;
    CopyStringToBuffer(Buffer, "hsl(%d, %d%%, %d%%)", Colour.Hue, Colour.Saturation, Colour.Lightness);
    return(Buffer->Location);
}

int
StringLength(char *String)
{
    int i = 0;
    while(String[i])
    {
        ++i;
    }
    return i;
}

char *
SanitisePunctuation(char *String)
{
    char *Ptr = String;
    while(*Ptr)
    {
        if(*Ptr == ' ')
        {
            *Ptr = '_';
        }
        if((*Ptr < '0' || *Ptr > '9') &&
                (*Ptr < 'a' || *Ptr > 'z') &&
                (*Ptr < 'A' || *Ptr > 'Z'))
        {
            *Ptr = '-';
        }
        ++Ptr;
    }
    return String;
}

int
BuildReference(ref_info *ReferencesArray, int RefIdentifier, int UniqueRefs, HMML_Reference Ref, HMML_Annotation Anno)
{
    if(Ref.url && Ref.title && Ref.author && Ref.publisher && Ref.isbn)
    {
        CopyString(ReferencesArray[UniqueRefs].ID, Ref.isbn);
        CopyString(ReferencesArray[UniqueRefs].Source, "%s (%s)", Ref.author, Ref.publisher);
        CopyString(ReferencesArray[UniqueRefs].RefTitle, Ref.title);
        CopyString(ReferencesArray[UniqueRefs].URL, Ref.url);
    }
    else if(Ref.page && Ref.url && Ref.title)
    {
        CopyString(ReferencesArray[UniqueRefs].ID, Ref.url);
        CopyString(ReferencesArray[UniqueRefs].Source, Ref.title);
        CopyString(ReferencesArray[UniqueRefs].RefTitle, Ref.page);
        CopyString(ReferencesArray[UniqueRefs].URL, Ref.url);
    }
    else if(Ref.site && Ref.page && Ref.url)
    {
        CopyString(ReferencesArray[UniqueRefs].ID, Ref.url);
        CopyString(ReferencesArray[UniqueRefs].Source, Ref.site);
        CopyString(ReferencesArray[UniqueRefs].RefTitle, Ref.page);
        CopyString(ReferencesArray[UniqueRefs].URL, Ref.url);
    }
    else if(Ref.site && Ref.url && Ref.title)
    {
        CopyString(ReferencesArray[UniqueRefs].ID, Ref.url);
        CopyString(ReferencesArray[UniqueRefs].Source, Ref.site);
        CopyString(ReferencesArray[UniqueRefs].RefTitle, Ref.title);
        CopyString(ReferencesArray[UniqueRefs].URL, Ref.url);
    }
    else if(Ref.title && Ref.author && Ref.isbn)
    {
        CopyString(ReferencesArray[UniqueRefs].ID, Ref.isbn);
        CopyString(ReferencesArray[UniqueRefs].Source, Ref.author);
        CopyString(ReferencesArray[UniqueRefs].RefTitle, Ref.title);
        //TODO(matt): Look into finding the best ISBN searcher the web has to offer
        CopyString(ReferencesArray[UniqueRefs].URL, "http://www.isbnsearch.org/isbn/%s", Ref.isbn);
    }
    else if(Ref.url && Ref.article && Ref.author)
    {
        CopyString(ReferencesArray[UniqueRefs].ID, Ref.url);
        CopyString(ReferencesArray[UniqueRefs].Source, Ref.author);
        CopyString(ReferencesArray[UniqueRefs].RefTitle, Ref.article);
        CopyString(ReferencesArray[UniqueRefs].URL, Ref.url);
    }
    else if(Ref.url && Ref.title && Ref.author)
    {
        CopyString(ReferencesArray[UniqueRefs].ID, Ref.url);
        CopyString(ReferencesArray[UniqueRefs].Source, Ref.author);
        CopyString(ReferencesArray[UniqueRefs].RefTitle, Ref.title);
        CopyString(ReferencesArray[UniqueRefs].URL, Ref.url);
    }
    else if(Ref.url && Ref.title)
    {
        CopyString(ReferencesArray[UniqueRefs].ID, Ref.url);
        CopyString(ReferencesArray[UniqueRefs].RefTitle, Ref.title);
        CopyString(ReferencesArray[UniqueRefs].URL, Ref.url);
    }
    else if(Ref.site && Ref.url)
    {
        CopyString(ReferencesArray[UniqueRefs].ID, Ref.url);
        CopyString(ReferencesArray[UniqueRefs].RefTitle, Ref.site);
        CopyString(ReferencesArray[UniqueRefs].URL, Ref.url);
    }
    else
    {
        return 1;
    }

    CopyString(ReferencesArray[UniqueRefs].Identifier[ReferencesArray[UniqueRefs].IdentifierCount].Timecode, Anno.time);
    ReferencesArray[UniqueRefs].Identifier[ReferencesArray[UniqueRefs].IdentifierCount].Identifier = RefIdentifier;
    return 0;
}

char *CategoryMedium[][3] =
{
    // category     icon        written name
    { "authored",   "&#128490", "Chat Comment"}, // TODO(matt): Conditionally handle Chat vs Guest Comments
    { "blackboard", "&#128396", "Blackboard"},
    { "owl",        "&#129417", "Owl of Shame"},
    { "default",    "&#128430", "Programming"}, // TODO(matt): Potentially make this configurable per project
    { "rant",       "&#128162", "Rant"},
    { "research",   "&#128214", "Research"},
    { "run",        "&#127939", "In-Game"} // TODO(matt): Potentially make this configurable per project
};

void
BuildFilter(category_info *CategoriesArray, int *UniqueCategories, char *Marker)
{
    int i = 0;
    for(i = 0; i < *UniqueCategories; ++i)
    {
        if(!StringsDiffer(Marker, CategoriesArray[i].Category))
        {
            return;
        }
    }

    ++*UniqueCategories;
    CopyString(CategoriesArray[i].Category, Marker);
    for(int j = 0; j < ArrayCount(CategoryMedium); ++j)
    {
        if(!StringsDiffer(CategoryMedium[j][0], Marker))
        {
            CategoriesArray[i].IsMedium = TRUE;
        }
    }
    // TODO(matt): Sort the CategoriesArray
}

void
BuildCategories(buffer *AnnotationClass, buffer *Category, int *MarkerIndex, bool *HasCategory, char *Marker)
{
    for(int i = 0; i < ArrayCount(CategoryMedium); ++i)
    {
        if(!StringsDiffer(CategoryMedium[i][0], Marker))
        {
            CopyStringToBuffer(AnnotationClass, " %s", SanitisePunctuation(Marker));
            ++*MarkerIndex;
            return;
        }
    }
    if(*HasCategory == FALSE)
    {
        CopyStringToBuffer(Category, "<span class=\"categories\">");
        *HasCategory = TRUE;
    }

    CopyStringToBuffer(Category, "<div class=\"category %s\"></div>",
        SanitisePunctuation(Marker));

    CopyStringToBuffer(AnnotationClass, " cat_%s",
            SanitisePunctuation(Marker));
    ++*MarkerIndex;
    return;
}

int
StringToInt(char *String)
{
    int Result = 0;
    while(*String)
    {
        Result = Result * 10 + (*String - '0');
        ++String;
    }
    return Result;
}

int
BuildQuote(quote_info *Info, char *Speaker, int ID)
{
    // TODO(matt): Pull these paths from the config
    //             Also notable that there are different paths for different projects
    //
    // TODO(matt): Handle csv's double quote escaping stuff

    char *QuoteDir = "/home/matt/git/GitHub/insofaras/25fc16d58a297a486334";
    if(!StringsDiffer(Speaker, "handmade_hero"))
    {
        QuoteDir = "/home/matt/git/GitHub/Chronister/01e754a09fd84c839ad7";
    }

    char Path[255] = {0};
    sprintf(Path, "%s/#%s", QuoteDir, Speaker);
    FILE *File;
    if(!(File = fopen(Path, "r")))
    {
        perror("hmml_to_html");
        return 1;
    }

    fseek(File, 0, SEEK_END);
    int Length = ftell(File);
    fseek(File, 0, SEEK_SET);
    char *Buffer;
    if(!(Buffer = malloc(Length)))
    {
        perror("hmml_to_html");
    }
    fread(Buffer, Length, 1, File);
    fclose(File);

    // TODO(matt): Search the quote store in reverse
    char *InPtr = Buffer;

    while(InPtr - Buffer < Length)
    {

        char InID[4] = { 0 };
        char *OutPtr = InID;
        while(*InPtr != ',')
        {
            *OutPtr++ = *InPtr++;
        }
        *OutPtr = '\0';

        if(StringToInt(InID) == ID)
        {
            InPtr += 2;
            OutPtr = Info->Date;
            while(*InPtr != '"')
            {
                *OutPtr++ = *InPtr++;
            }
            *OutPtr = '\0';

            InPtr += 3;
            OutPtr = Info->Text;
            while(*InPtr != '\n')
            {
                if(*InPtr == '\\')
                {
                    ++InPtr;
                }
                *OutPtr++ = *InPtr++;
            }
            *--OutPtr = '\0';

            free(Buffer);
            return 0;
        }
        else
        {
            while(*InPtr != '\n')
            {
                ++InPtr;
            }
            ++InPtr;
        }
    }

    free(Buffer);
    return 1;
}

void
GenerateTopicColours(buffer *Colour, char *Topic)
{
    for(int i = 0; i < ArrayCount(CategoryMedium); ++i)
    {
        if(!StringsDiffer(Topic, CategoryMedium[i][0]))
        {
            return;
        }
    }

    FILE *TopicsFile;
    char *TopicsBuffer;
    // TODO(matt): Consider (optionally) pulling this path from the config
    if((TopicsFile = fopen("topics.css", "a+")))
    {
        fseek(TopicsFile, 0, SEEK_END);
        int TopicsLength = ftell(TopicsFile);
        fseek(TopicsFile, 0, SEEK_SET);

        // TODO(matt): May this not just ClaimBuffer?
        if(!(TopicsBuffer = malloc(TopicsLength)))
        {
            perror("GenerateTopicColours");
            return;
        }

        fread(TopicsBuffer, TopicsLength, 1, TopicsFile);

        char *TopicsPtr = TopicsBuffer;

        while(TopicsPtr - TopicsBuffer < TopicsLength)
        {
            TopicsPtr += StringLength(".category.");
            if(!StringsDifferL(SanitisePunctuation(Topic), TopicsPtr, StringLength(Topic)))
            {
                free(TopicsBuffer);
                fclose(TopicsFile);
                return;
            }
            while(TopicsPtr - TopicsBuffer < TopicsLength && *TopicsPtr != '\n')
            {
                ++TopicsPtr;
            }
            ++TopicsPtr;
        }

        fprintf(TopicsFile, ".category.%s { border: 1px solid %s; background: %s; }\n",
            SanitisePunctuation(Topic), StringToColourHash(Colour, Topic), StringToColourHash(Colour, Topic));

        fclose(TopicsFile);
        free(TopicsBuffer);
    }
    else
    {
        perror("GenerateTopicColours");
        return;
    }
}

#define CONFIG 0

#if CONFIG
typedef struct
{
    char *Username;
    char *Display_Name;
    char *Homepage;
    char *Funding_Platform;
    char *Funding_Username;
    unsigned int Index;
} credentials;

typedef struct
{
    credentials Credentials;
} config;

int
ParseConfig(buffer *Buffer, char *Username)
{
    /*
       Essentially, I want to pass a Username to this, and have it write the credentials into the Config buffer
       Let's start by just grabbing the stuff and printing it out
     */

    // TODO(matt): Actually figure out the "standard" config location
    char Config_Location[255];
    if(getenv("XDG_CONFIG_HOME"))
    {
        sprintf(Config_Location, "%s/hmml.conf", getenv("XDG_CONFIG_HOME"));
    }
    else if(getenv("HOME"))
    {
        sprintf(Config_Location, "%s/.config/hmml.conf", getenv("HOME"));
    }
    else
    {
        fprintf(stderr, "Config file location not set");
        return 1;
    }

    FILE *InFile;
    if(!(InFile = fopen(Config_Location, "r")))
    {
        perror(Config_Location);
        return 2;
    }

    printf("Reading: %s\n", Config_Location);

    fseek(InFile, 0, SEEK_END);
    int InSize = ftell(InFile);
    fseek(InFile, 0, SEEK_SET);

    char *InBuffer;
    //config Config = { 0 };

    if(!(InBuffer = malloc(InSize)))
    {
        perror("ParseConfig");
        return 3;
    }

    fread(InBuffer, InSize, 1, InFile);
    fclose(InFile);

    char *InPtr = InBuffer;
    char OutBuffer[256];
    //char *OutPtr = Config.Credentials.Display_Name;
    char *OutPtr = OutBuffer;
    bool Quoted = FALSE;
    bool FoundCredentials, ParsingUsername = FALSE;
    unsigned int ScopeDepth = 0;

    while(InPtr - InBuffer < InSize)
    {
        switch(*InPtr)
        {
            case '#':
                {
                    if(!Quoted)
                    {
                        printf("    We are commenting\n");
                        while(InPtr - InBuffer < InSize && *InPtr != '\n')
                        {
                            ++InPtr;
                        }
                        ++InPtr;
                        while(InPtr - InBuffer < InSize && (*InPtr == ' ' || *InPtr == '\n'))
                        {
                            ++InPtr;
                        }
                    }
                    else
                    {
                        *OutPtr++ = *InPtr++;
                    }
                    break;
                }
            case '{':
                {
                    if(!Quoted)
                    {
                        ++ScopeDepth;
                        ++InPtr;
                        while(*InPtr == '\n' || *InPtr == ' ')
                        {
                            ++InPtr;
                        }
                        printf("    We have entered a scope\n");
                    }
                    else
                    {
                        *OutPtr++ = *InPtr++;
                    }
                    break;
                }
            case '}':
                {
                    if(!Quoted)
                    {
                        --ScopeDepth;
                        ++InPtr;
                        printf("    We have left a scope\n");
                    }
                    else
                    {
                        *OutPtr++ = *InPtr++;
                    }
                    break;
                }
#if 1
            case ' ':
                {
                    if(!Quoted)
                    {
                        ++InPtr;
                        *OutPtr = '\0';
                        OutPtr = OutBuffer;
                        printf("%s\n", OutBuffer);

                        // TODO(matt): Switch on the OutBuffer? I have a feeling that isn't actually possible, though
                        if(!StringsDiffer("credentials", OutBuffer))
                        {
                            FoundCredentials = TRUE;
                            printf("    We have found the credentials block\n");
                        }
                        if(ParsingUsername)
                        {
                            printf("    The username is %s\n", OutBuffer);
                            ParsingUsername = FALSE;
                        }
                        if(FoundCredentials && (!StringsDiffer("username", OutBuffer)))
                        {
                            ParsingUsername = TRUE;
                            printf("    We have found the username\n");
                        }
                    }
                    else
                    {
                        *OutPtr++ = *InPtr++;
                    }
                    break;
                }
#endif
            case '"':
                {
                    if(!Quoted)
                    {
                        Quoted = TRUE;
                        printf("    We are quoting!\n");
                    }
                    else
                    {
                        Quoted = FALSE;
                        printf("    We are no longer quoting!\n");
                    }
                    ++InPtr;
                    break;
                }
            case ';':
                {
                    if(!Quoted)
                    {
                        printf("    We have reached the end of a setting\n");
                        ++InPtr;
                    }
                    else
                    {
                        *OutPtr++ = *InPtr++;
                    }
                }
            case '\n':
                {
                    if(!Quoted)
                    {
                        if(InPtr - InBuffer < InSize)
                        {
                            *OutPtr = '\0';
                            OutPtr = OutBuffer;
                            // TODO(matt)
                            if(!StringsDiffer("credentials", OutBuffer))
                            {
                                FoundCredentials = TRUE;
                                printf("    We have found the credentials block\n");
                            }
                            if(ParsingUsername)
                            {
                                printf("    The username is %s\n", OutBuffer);
                                ParsingUsername = FALSE;
                            }
                            if(FoundCredentials && (!StringsDiffer("username", OutBuffer)))
                            {
                                ParsingUsername = TRUE;
                                printf("    We have found the username\n");
                            }
                            printf("%s\n", OutBuffer);
                            ++InPtr;
                            while(InPtr - InBuffer < InSize && *InPtr == ' ') // NOTE(matt): Skip indentation whitespace
                            {
                                ++InPtr;
                            }
                        }
                    }
                    else
                    {
                        *OutPtr++ = *InPtr++;
                    }

                    if(InPtr - InBuffer == InSize)
                    {
                        printf("    We have reached the EOF\n");
                    }
                    break;
                }
            default:
                {
                    *OutPtr++ = *InPtr++;
                    break;
                }
        }
    }


#if 0
    while(InPtr - InBuffer < InSize)
    {
        while(*InPtr != '\n')
        {
            *OutPtr++ = *InPtr++;
        }
        *OutPtr = '\0';
        printf("%s\n", Config.Credentials.Display_Name);
    }
#endif

    free(InBuffer);

    // Reading from the config file, parsing it inline (on the stack) and writing into the buffer *Config

    return 0;
}

#endif

int
main(int ArgC, char **Args)
{
    if(ArgC < 2)
    {
        fprintf(stderr, "Usage: %s filename(s)\n", Args[0]);
        return 1;
    }
    
    // NOTE(matt): Init MemoryArena
    char *MemoryArena;
    int ArenaSize = 1024 * 1024 * 4;
    if(!(MemoryArena = calloc(ArenaSize, 1)))
    {
        perror(Args[0]);
        return 1;
    }
    int ClaimedMemory = 0;
    
    // NOTE(matt): Setup buffers and ptrs
    char *InPtr;

#if CONFIG
    buffer Config;
#endif

    // NOTE(matt): Tree structure of buffer dependencies
    // Master
    //     Title
    //         QuoteMenu
    //         ReferenceMenu
    //         FilterMenu
    //             FilterTopics
    //             FilterMedia
    //     Player
    //         Colour
    //         Annotation
    //             AnnotationHeader
    //                 AnnotationClass
    //                 AnnotationData
    //             Text
    //                 Category
    //     FilterState

    buffer Master;

    buffer Title;
    buffer QuoteMenu;
    buffer ReferenceMenu;
    buffer FilterMenu;
    buffer FilterTopics;
    buffer FilterMedia;

    buffer Player;
    buffer Colour;
    buffer Annotation;
    buffer AnnotationHeader;
    buffer AnnotationClass;
    buffer AnnotationData;
    buffer Text;
    buffer Category;

    buffer FilterState;

    for(int FileIndex = 1; FileIndex < ArgC; ++FileIndex)
    {
        FILE *InFile;
        if(!(InFile = fopen(Args[FileIndex], "r")))
        {
            perror(Args[0]);
            free(MemoryArena);
            return 1;
        }

#if CONFIG
        ClaimBuffer(MemoryArena, &ClaimedMemory, &Config, "Config", 1024);
#endif

        HMML_Output HMML = hmml_parse_file(InFile);
        fclose(InFile);

        if(HMML.well_formed)
        {
            // NOTE(matt): Tree structure of "global" buffer dependencies
            // Master
            //     Title
            //         QuoteMenu
            //         ReferenceMenu
            //         FilterMenu
            //             FilterTopics
            //             FilterMedia
            //     Player
            //         Colour
            //         Annotation
            //     FilterState

            ClaimBuffer(MemoryArena, &ClaimedMemory, &Master, "Master", 1024 * 512);

            ClaimBuffer(MemoryArena, &ClaimedMemory, &Title, "Title", 1024 * 16);
            ClaimBuffer(MemoryArena, &ClaimedMemory, &QuoteMenu, "QuoteMenu", 1024 * 16);
            ClaimBuffer(MemoryArena, &ClaimedMemory, &ReferenceMenu, "ReferenceMenu", 1024 * 16);
            ClaimBuffer(MemoryArena, &ClaimedMemory, &FilterMenu, "FilterMenu", 1024 * 16);
            ClaimBuffer(MemoryArena, &ClaimedMemory, &FilterTopics, "FilterTopics", 1024 * 8);
            ClaimBuffer(MemoryArena, &ClaimedMemory, &FilterMedia, "FilterMedia", 1024 * 8);

            ClaimBuffer(MemoryArena, &ClaimedMemory, &Player, "Player", 1024 * 256);
            ClaimBuffer(MemoryArena, &ClaimedMemory, &Colour, "Colour", 32);
            ClaimBuffer(MemoryArena, &ClaimedMemory, &Annotation, "Annotation", 1024 * 8);

            ClaimBuffer(MemoryArena, &ClaimedMemory, &FilterState, "FilterState", 1024 * 4);

            ref_info ReferencesArray[200] = { 0 };
            category_info CategoriesArray[64] = { 0 };

            bool HasQuoteMenu = FALSE;
            bool HasReferenceMenu = FALSE;
            bool HasFilterMenu = FALSE;

            int QuoteIdentifier = 0x3b1;
            int RefIdentifier = 1;
            int UniqueRefs = 0;
            int UniqueCategories = 0;

            CopyStringToBuffer(&Title,
"        <div class=\"title %s\">\n"
"            <span class=\"episode_name\">%s</span>\n", HMML.metadata.project, HMML.metadata.title);

            CopyStringToBuffer(&Player,
"        <div class=\"player_container\">\n"
"            <div class=\"video_container\" data-videoId=\"%s\"></div>\n"
"            <div class=\"markers_container %s\">\n", HMML.metadata.id, HMML.metadata.project);

#if DEBUG
                printf(" --- Entering Annotations Loop ---\n");
#endif

            for(int AnnotationIndex = 0; AnnotationIndex < HMML.annotation_count; ++AnnotationIndex)
            {
#if DEBUG
                printf("%d\n", AnnotationIndex);
#endif
                HMML_Annotation *Anno = HMML.annotations + AnnotationIndex;
                bool HasCategory = FALSE;
                bool HasQuote = FALSE;
                bool HasReference = FALSE;

                quote_info QuoteInfo = { 0 };

                // NOTE(matt): Tree structure of "annotation local" buffer dependencies
                // AnnotationHeader
                //     AnnotationClass
                //     AnnotationData
                // Text
                //     Category

                ClaimBuffer(MemoryArena, &ClaimedMemory, &AnnotationHeader, "AnnotationHeader", 512);
                ClaimBuffer(MemoryArena, &ClaimedMemory, &AnnotationClass, "AnnotationClass", 256);
                ClaimBuffer(MemoryArena, &ClaimedMemory, &AnnotationData, "AnnotationData", 128);
                ClaimBuffer(MemoryArena, &ClaimedMemory, &Text, "Text", 1024 * 4);
                ClaimBuffer(MemoryArena, &ClaimedMemory, &Category, "Category", 256);

                CopyStringToBuffer(&AnnotationHeader,
"                <div data-timestamp=\"%d\"",
TimecodeToSeconds(Anno->time));

                CopyStringToBuffer(&AnnotationClass,
" class=\"marker");

                if(Anno->author)
                {
                    if(!HasFilterMenu)
                    {
                        HasFilterMenu = TRUE;
                    }
                    BuildFilter(CategoriesArray, &UniqueCategories, "authored");
                    CopyStringToBuffer(&AnnotationClass, " authored");
                    CopyStringToBuffer(&Text,
"<span class=\"author\" style=\"color: %s;\">%s</span> ",
StringToColourHash(&Colour, Anno->author),
Anno->author);
                }

                InPtr = Anno->text;

                int MarkerIndex = 0, RefIndex = 0;
                while(*InPtr || RefIndex < Anno->reference_count)
                {
                    if(MarkerIndex < Anno->marker_count &&
                            InPtr - Anno->text == Anno->markers[MarkerIndex].offset)
                    {
                        char *Readable = Anno->markers[MarkerIndex].parameter
                            ? Anno->markers[MarkerIndex].parameter
                            : Anno->markers[MarkerIndex].marker;
                        if(Anno->markers[MarkerIndex].type == HMML_MEMBER)
                        {
                            CopyStringToBuffer(&Text,
                                    // TODO(matt): Hoverbox
                                    //             We should get instructions on how to get this info in the config
"<a href=\"https://handmade.network/m/%s\" target=\"blank\" style=\"color: %s; text-decoration: none\">%.*s</a>",
Anno->markers[MarkerIndex].marker,
StringToColourHash(&Colour, Anno->markers[MarkerIndex].marker),
StringLength(Readable), InPtr);
                            InPtr += StringLength(Readable);
                            ++MarkerIndex;
                        }
                        else if(Anno->markers[MarkerIndex].type == HMML_PROJECT)
                        {
                            CopyStringToBuffer(&Text,
                                    // TODO(matt): Hoverbox
                                    //             We should get instructions on how to get this info in the config
"<a href=\"https://%s.handmade.network/\" target=\"blank\" style=\"color: %s; text-decoration: none\">%s</a>",
Anno->markers[MarkerIndex].marker,
StringToColourHash(&Colour, Anno->markers[MarkerIndex].marker),
Readable);
                            InPtr += StringLength(Readable);
                            ++MarkerIndex;
                        }
                        else if(Anno->markers[MarkerIndex].type == HMML_CATEGORY)
                        {
                            GenerateTopicColours(&Colour, Anno->markers[MarkerIndex].marker);
                            // TODO(matt): Maybe stuff this into BuildCategories
                            if(!HasFilterMenu)
                            {
                                HasFilterMenu = TRUE;
                            }
                            BuildFilter(CategoriesArray, &UniqueCategories, Anno->markers[MarkerIndex].marker);
                            BuildCategories(&AnnotationClass, &Category, &MarkerIndex, &HasCategory, Anno->markers[MarkerIndex].marker);
                        }
                    }
                    
                    if(RefIndex < Anno->reference_count &&
                            InPtr - Anno->text == Anno->references[RefIndex].offset)
                    {
                        HMML_Reference *CurrentRef = Anno->references + RefIndex;
                        if(!HasReferenceMenu)
                        {
                            CopyStringToBuffer(&ReferenceMenu,
"            <div class=\"menu\">\n"
"                <span>References &#9660;</span>\n"
"                <div class=\"mouse_catcher\"></div>\n"
"                <div class=\"refs\">\n");

                            if(BuildReference(ReferencesArray, RefIdentifier, UniqueRefs, *CurrentRef, *Anno) == 1)
                            {
                                fprintf(stderr, "%s:%d: Cannot process new combination of reference info\n", Args[FileIndex], Anno->line);
                                hmml_free(&HMML);
                                free(MemoryArena);
                                return 1;
                            }
                            ++ReferencesArray[RefIdentifier - 1].IdentifierCount;
                            ++UniqueRefs;

                            HasReferenceMenu = TRUE;
                        }
                        else
                        {
                            for(int i = 0; i < UniqueRefs; ++i)
                            {
                                if(CurrentRef->isbn)
                                {
                                    if(!StringsDiffer(CurrentRef->isbn, ReferencesArray[i].ID))
                                    {
                                        CopyString(ReferencesArray[i].Identifier[ReferencesArray[i].IdentifierCount].Timecode, Anno->time);
                                        ReferencesArray[i].Identifier[ReferencesArray[i].IdentifierCount].Identifier = RefIdentifier;
                                        ++ReferencesArray[i].IdentifierCount;
                                        goto AppendedIdentifier;
                                    }
                                }
                                else if(CurrentRef->url)
                                {
                                    if(!StringsDiffer(CurrentRef->url, ReferencesArray[i].ID))
                                    {
                                        CopyString(ReferencesArray[i].Identifier[ReferencesArray[i].IdentifierCount].Timecode, Anno->time);
                                        ReferencesArray[i].Identifier[ReferencesArray[i].IdentifierCount].Identifier = RefIdentifier;
                                        ++ReferencesArray[i].IdentifierCount;
                                        goto AppendedIdentifier;
                                    }
                                }
                                else
                                {
                                    fprintf(stderr, "%s:%d: Reference must have an ISBN or URL\n", Args[FileIndex], Anno->line);
                                    hmml_free(&HMML);
                                    free(MemoryArena);
                                    return 1;
                                }
                            }

                            if(BuildReference(ReferencesArray, RefIdentifier, UniqueRefs, *CurrentRef, *Anno) == 1)
                            {
                                fprintf(stderr, "%s:%d: Cannot process new combination of reference info\n", Args[FileIndex], Anno->line);
                                hmml_free(&HMML);
                                free(MemoryArena);
                                return 1;
                            }
                            ++ReferencesArray[UniqueRefs].IdentifierCount;
                            ++UniqueRefs;
                        }
AppendedIdentifier:
                        if(!HasReference)
                        {
                            if(CurrentRef->isbn)
                            {
                                CopyStringToBuffer(&AnnotationData, " data-ref=\"%s", CurrentRef->isbn);
                            }
                            else if(CurrentRef->url)
                            {
                                CopyStringToBuffer(&AnnotationData, " data-ref=\"%s", CurrentRef->url);
                            }
                            else
                            {
                                fprintf(stderr, "%s:%d: Reference must have an ISBN or URL\n", Args[FileIndex], Anno->line);
                                hmml_free(&HMML);
                                free(MemoryArena);
                                return 1;
                            }

                            HasReference = TRUE;
                        }
                        else
                        {
                            if(CurrentRef->isbn)
                            {
                                CopyStringToBuffer(&AnnotationData, ",%s", CurrentRef->isbn);
                            }
                            else if(CurrentRef->url)
                            {
                                CopyStringToBuffer(&AnnotationData, ",%s", CurrentRef->url);
                            }
                            else
                            {
                                fprintf(stderr, "%s:%d: Reference must have an ISBN or URL", Args[FileIndex], Anno->line);
                                hmml_free(&HMML);
                                free(MemoryArena);
                                return 1;
                            }
                        }

                        CopyStringToBuffer(&Text, "<sup>%d</sup>", RefIdentifier);

                        ++RefIndex;
                        ++RefIdentifier;
                    }

                    if(*InPtr)
                    {
                        *Text.Ptr++ = *InPtr++;
                    }
                }

                if(Anno->is_quote)
                {
                    if(!HasQuoteMenu)
                    {
                        CopyStringToBuffer(&QuoteMenu,
"            <div class=\"menu\">\n"
"                <span>Quotes &#9660;</span>\n"
"                <div class=\"mouse_catcher\"></div>\n"
"                <div class=\"refs\">\n");

                        HasQuoteMenu = TRUE;
                    }

                    if(!HasReference)
                    {
                        CopyStringToBuffer(&AnnotationData, " data-ref=\"&#%d;", QuoteIdentifier);
                    }
                    else
                    {
                        CopyStringToBuffer(&AnnotationData, ",&#%d;", QuoteIdentifier);
                    }

                    HasQuote = TRUE;


                    if(BuildQuote(&QuoteInfo, HMML.metadata.twitch ? HMML.metadata.twitch : HMML.metadata.member, Anno->quote.id) == 1)
                    {
                        fprintf(stderr, "%s:%d: Quote #%s %d not found! Consider pulling the latest quotes\n",
                                Args[FileIndex],
                                Anno->line,
                                HMML.metadata.twitch ? HMML.metadata.twitch : HMML.metadata.member,
                                Anno->quote.id);
                        hmml_free(&HMML);
                        free(MemoryArena);
                        return 1;
                    }

                    CopyStringToBuffer(&QuoteMenu,
"                    <span data-id=\"&#%d;\" class=\"ref\">\n"
"                        <span class=\"ref_content\">\n"
"                            <div class=\"source\">Quote %d</div>\n"
"                            <div class=\"ref_title\">%s</div>\n"
"                            <div class=\"quote_byline\">&mdash;%s, %s</div>\n"
"                        </span>\n"
"                        <div class=\"ref_indices\">\n"
"                            <span data-timestamp=\"%d\" class=\"timecode\"><span class=\"ref_index\">[&#%d;]</span><span class=\"time\">%s</span></span>\n"
"                        </div>\n"
"                    </span>\n",
QuoteIdentifier,
Anno->quote.id,
QuoteInfo.Text,
HMML.metadata.twitch ? HMML.metadata.twitch : HMML.metadata.member,
QuoteInfo.Date,
TimecodeToSeconds(Anno->time),
QuoteIdentifier,
Anno->time);
                    if(!Anno->text[0])
                    {
                        CopyStringToBuffer(&Text, "&#8220;%s&#8221;", QuoteInfo.Text);
                    }
                    CopyStringToBuffer(&Text, "<sup>&#%d;</sup>", QuoteIdentifier);
                    ++QuoteIdentifier;
                }

                while(MarkerIndex < Anno->marker_count)
                {
                    GenerateTopicColours(&Colour, Anno->markers[MarkerIndex].marker);
                    // TODO(matt): Maybe stuff this into BuildCategories
                    if(!HasFilterMenu)
                    {
                        HasFilterMenu = TRUE;
                    }
                    if(Anno->markers[MarkerIndex].marker)
                    {
                        BuildFilter(CategoriesArray, &UniqueCategories, Anno->markers[MarkerIndex].marker);
                    }
                    BuildCategories(&AnnotationClass, &Category, &MarkerIndex, &HasCategory, Anno->markers[MarkerIndex].marker);
                }

                CopyStringToBuffer(&AnnotationClass, "\"");
                CopyBuffer(&AnnotationHeader, &AnnotationClass);

                if(HasQuote || HasReference)
                {
                    CopyStringToBuffer(&AnnotationData, "\"");
                    CopyBuffer(&AnnotationHeader, &AnnotationData);
                }
                CopyStringToBuffer(&AnnotationHeader, ">\n");

                CopyBuffer(&Annotation, &AnnotationHeader);
                CopyStringToBuffer(&Annotation,
"                    <div class=\"content\"><span class=\"timecode\">%s</span>",
Anno->time);

#if 0
                // TODO(matt): Handle special-cases, i.e. default media, and possibly other things
                if(!HasCategory)
                {
                    BuildFilter(CategoriesArray, &UniqueCategories, "default");
                    BuildCategories(&AnnotationClass, &Category, &MarkerIndex, &HasCategory, "default");
                }
#endif

                if(HasCategory)
                {
                    CopyStringToBuffer(&Category, "</span>");
                    CopyBuffer(&Text, &Category);
                }

                *Text.Ptr = '\0';

                CopyBuffer(&Annotation, &Text);

                CopyStringToBuffer(&Annotation, "</div>\n"
"                    <div class=\"progress faded\">\n"
"                        <div class=\"content\"><span class=\"timecode\">%s</span>",
Anno->time);

                CopyBuffer(&Annotation, &Text);

                CopyStringToBuffer(&Annotation, "</div>\n"
"                    </div>\n"
"                    <div class=\"progress main\">\n"
"                        <div class=\"content\"><span class=\"timecode\">%s</span>",
Anno->time);

                CopyBuffer(&Annotation, &Text);

                CopyStringToBuffer(&Annotation, "</div>\n"
"                    </div>\n"
"                </div>\n");

                CopyBuffer(&Player, &Annotation);

                // NOTE(matt): Tree structure of "annotation local" buffer dependencies
                //     Category
                // Text
                //     AnnotationData
                //     AnnotationClass
                // AnnotationHeader

                DeclaimBuffer(&Category, &ClaimedMemory);
                DeclaimBuffer(&Text, &ClaimedMemory);
                DeclaimBuffer(&AnnotationData, &ClaimedMemory);
                DeclaimBuffer(&AnnotationClass, &ClaimedMemory);
                DeclaimBuffer(&AnnotationHeader, &ClaimedMemory);
                Annotation.Ptr = Annotation.Location;
            }

#if DEBUG
                printf(" --- End of Annotations Loop ---\n\n");
#endif
            if(HasQuoteMenu)
            {
                CopyStringToBuffer(&QuoteMenu,
"                </div>\n"
"            </div>\n");
                CopyBuffer(&Title, &QuoteMenu);
            }

            if(HasReferenceMenu)
            {
                for(int i = 0; i < UniqueRefs; ++i)
                {
                        CopyStringToBuffer(&ReferenceMenu,
"                    <a data-id=\"%s\" href=\"%s\" target=\"_blank\" class=\"ref\">\n"
"                        <span class=\"ref_content\">\n",
ReferencesArray[i].ID,
ReferencesArray[i].URL);

                        if(*ReferencesArray[i].Source)
                        {
                            CopyStringToBuffer(&ReferenceMenu,
"                            <div class=\"source\">%s</div>\n"
"                            <div class=\"ref_title\">%s</div>\n",
ReferencesArray[i].Source,
ReferencesArray[i].RefTitle);
                        }
                        else
                        {
                            CopyStringToBuffer(&ReferenceMenu,
"                            <div class=\"ref_title\">%s</div>\n",
ReferencesArray[i].RefTitle);
                        }
                            CopyStringToBuffer(&ReferenceMenu,
"                        </span>\n");

                        for(int j = 0; j < ReferencesArray[i].IdentifierCount;)
                        {
                            CopyStringToBuffer(&ReferenceMenu,
"                        <div class=\"ref_indices\">\n                            ");
                            for(int k = 0; k < 3 && j < ReferencesArray[i].IdentifierCount; ++k, ++j)
                            {
                                CopyStringToBuffer(&ReferenceMenu,
"<span data-timestamp=\"%d\" class=\"timecode\"><span class=\"ref_index\">[%d]</span><span class=\"time\">%s</span></span>",
TimecodeToSeconds(ReferencesArray[i].Identifier[j].Timecode),
ReferencesArray[i].Identifier[j].Identifier,
ReferencesArray[i].Identifier[j].Timecode);
                            }
                        CopyStringToBuffer(&ReferenceMenu, "\n"
"                        </div>\n");
                        }

                        CopyStringToBuffer(&ReferenceMenu,
"                    </a>\n");
                }

                CopyStringToBuffer(&ReferenceMenu,
"                </div>\n"
"            </div>\n");
                CopyBuffer(&Title, &ReferenceMenu);
            }

            if(HasFilterMenu)
            {
                CopyStringToBuffer(&FilterState, "var filterState = {\n");
                for(int i = 0; i < UniqueCategories; ++i)
                {
                    CopyStringToBuffer(&FilterState, "\"%s\":\t{ \"type\": \"%s\",\t\"off\": false },\n",
                            CategoriesArray[i].Category, CategoriesArray[i].IsMedium ? "medium" : "topic");
                }
                CopyStringToBuffer(&FilterState, "};\n"
                        "\n");

                CopyStringToBuffer(&FilterMenu,
"            <div class=\"menu filter\">\n"
"                <span><img src=\"hues_HCL.png\"></span>\n"
"                <div class=\"filter_container\">\n"
"                    <div class=\"filter_mode inclusive\">Filter mode: </div>\n"
"                    <div class=\"filters\">\n");

                {
                    bool HasTopic = FALSE;
                    bool HasMedium = FALSE;

                    for(int i = 0; i < UniqueCategories; ++i)
                    {
                        if(CategoriesArray[i].IsMedium)
                        {
                            if(!HasMedium)
                            {
                                CopyStringToBuffer(&FilterMedia,
"                        <div class=\"filter_media\">\n"
"                            <div class=\"filter_title\">Media</div>\n");

                                HasMedium = TRUE;
                            }

                            int j;
                            for(j = 0; j < ArrayCount(CategoryMedium); ++j)
                            {
                                if(!StringsDiffer(CategoriesArray[i].Category, CategoryMedium[j][0]))
                                {
                                    break;
                                }
                            }

                            CopyStringToBuffer(&FilterMedia,
"                            <div class=\"filter_content %s\">\n"
"                                <span class=\"icon\">%s</span><span class=\"text\">%s</span>\n"
"                            </div>\n",
CategoriesArray[i].Category,
CategoryMedium[j][1],
CategoryMedium[j][2]
);

                            /*
                        </div>
                             */
                        }
                        else
                        {
                            if(!HasTopic)
                            {
                                CopyStringToBuffer(&FilterMenu,
"                        <div class=\"filter_topics\">\n"
"                            <div class=\"filter_title\">Topics</div>\n");

                                HasTopic = TRUE;
                            }
                        CopyStringToBuffer(&FilterTopics,
"                            <div class=\"filter_content %s\">\n"
"                                <span class=\"icon category %s\"></span><span class=\"text\">%s</span>\n"
"                            </div>\n",
CategoriesArray[i].Category,
CategoriesArray[i].Category,
CategoriesArray[i].Category);
                        }
                    }

                    if(HasTopic)
                    {
                        CopyStringToBuffer(&FilterTopics,
"                        </div>\n");
                        CopyBuffer(&FilterMenu, &FilterTopics);
                    }
                    if(HasMedium)
                    {
                        CopyStringToBuffer(&FilterMedia,
"                        </div>\n");
                        CopyBuffer(&FilterMenu, &FilterMedia);
                    }
                }


                 CopyStringToBuffer(&FilterMenu,
"                    </div>\n"
"                </div>\n"
"            </div>\n");
            }

            CopyBuffer(&Title, &FilterMenu);

#if CONFIG
            // TODO(matt): Here is where I test ParseConfig
            ParseConfig(&Config, HMML.metadata.annotator);
#endif
            CopyStringToBuffer(&Title,
"            <span class=\"annotator_container\">Annotator: <span class=\"annotator\">%s</span></span>\n"
"        </div>\n",
HMML.metadata.annotator);

            CopyStringToBuffer(&Player,
"            </div>\n"
"        </div>\n");

            //NOTE(matt): Collate the buffers!
#if DEBUG
            printf("Buffer Collation\n\n");
#endif

            CopyStringToBuffer(&Master,
"<html>\n"
"    <head>\n"
"        <meta charset=\"UTF-8\">\n"
"\n"
"        <!-- Load the player -->\n"
"        <script type=\"text/javascript\" src=\"player.js\"></script>\n"
"        <link rel=\"stylesheet\" type=\"text/css\" href=\"style.css\">\n"
"        <link rel=\"stylesheet\" type=\"text/css\" href=\"%s.css\">\n"
"        <link rel=\"stylesheet\" type=\"text/css\" href=\"topics.css\">\n"
"    </head>\n"
"    <body>\n", HMML.metadata.project);

            //NOTE(matt): Here is where we do all our CopyBuffer() calls
            CopyBuffer(&Master, &Title);
            CopyBuffer(&Master, &Player);
            //

            CopyStringToBuffer(&Master,
"        <script>\n"
"            var player = new Player(document.querySelector(\".player_container\"), onRefChanged);\n"
"            window.addEventListener(\"resize\", function() { player.updateSize(); });\n"
"                document.addEventListener(\"keypress\", function(ev) {\n"
"                    switch (ev.key) {\n"
"                        case 'n':\n"
"                        case 'd':\n"
"                        case 's': {\n"
"                                player.jumpToNextMarker();\n"
"                        } break;\n"
"\n"
"                        case 'p':\n"
"                        case 'a':\n"
"                        case 'w': {\n"
"                                player.jumpToPrevMarker();\n"
"                        } break;\n"
"    }\n"
"});\n"
"\n"
"var refTimecodes = document.querySelectorAll(\".refs .ref .timecode\");\n"
"for (var i = 0; i < refTimecodes.length; ++i) {\n"
"    refTimecodes[i].addEventListener(\"click\", function(ev) {\n"
"        if (player) {\n"
"            var time = ev.currentTarget.getAttribute(\"data-timestamp\");\n"
"            player.setTime(parseInt(time, 10));\n"
"            player.play();\n"
"            ev.preventDefault();\n"
"            ev.stopPropagation();\n"
"            return false;\n"
"        }\n"
"    });\n"
"}\n"
"\n"
"var filter = document.querySelector(\".filter\");\n"
"var filterModeElement = filter.querySelector(\".filter_mode\");\n"
"var filterMode = filterModeElement.classList[1];\n");

            if(HasFilterMenu)
            {
                CopyBuffer(&Master, &FilterState);
            }

            // NOTE(matt): Tree structure of "global" buffer dependencies
            //  FilterState
            //      Annotation
            //      Colour
            //  Player
            //          FilterMedia
            //          FilterTopics
            //      FilterMenu
            //      ReferenceMenu
            //      QuoteMenu
            //  Title

            DeclaimBuffer(&FilterState, &ClaimedMemory);
            DeclaimBuffer(&Annotation, &ClaimedMemory);
            DeclaimBuffer(&Colour, &ClaimedMemory);
            DeclaimBuffer(&Player, &ClaimedMemory);

            DeclaimBuffer(&FilterMedia, &ClaimedMemory);
            DeclaimBuffer(&FilterTopics, &ClaimedMemory);
            DeclaimBuffer(&FilterMenu, &ClaimedMemory);
            DeclaimBuffer(&ReferenceMenu, &ClaimedMemory);
            DeclaimBuffer(&QuoteMenu, &ClaimedMemory);
            DeclaimBuffer(&Title, &ClaimedMemory);

            CopyStringToBuffer(&Master,
"// Filter Mode Toggle\n"
"var testMarkers = document.querySelectorAll(\".marker\");\n"
"filterModeElement.addEventListener(\"click\", function(ev) {\n"
"    if(filterMode == \"inclusive\")\n"
"    {\n"
"        filterModeElement.classList.remove(\"inclusive\");\n"
"        filterModeElement.classList.add(\"exclusive\");\n"
"        filterMode = \"exclusive\";\n"
"\n"
"        for(var i = 0; i < testMarkers.length; ++i)\n"
"        {\n"
"            var testCategories = testMarkers[i].classList;\n"
"            for(var j = 0; j < testCategories.length; ++j)\n"
"            {\n"
"                if((testCategories[j].startsWith(\"off_\")) && !testMarkers[i].classList.contains(\"skip\"))\n"
"                {\n"
"                    testMarkers[i].classList.add(\"skip\");\n"
"                }\n"
"            }\n"
"        }\n"
"    }\n"
"    else\n"
"    {\n"
"        filterModeElement.classList.remove(\"exclusive\");\n"
"        filterModeElement.classList.add(\"inclusive\");\n"
"        filterMode = \"inclusive\";\n"
"\n"
"        for(var i = 0; i < testMarkers.length; ++i)\n"
"        {\n"
"            var testCategories = testMarkers[i].classList;\n"
"            for(var j = 0; j < testCategories.length; ++j)\n"
"            {\n"
"                if((testCategories[j] in filterState || testCategories[j].startsWith(\"cat_\")) && testMarkers[i].classList.contains(\"skip\"))\n"
"                {\n"
"                    testMarkers[i].classList.remove(\"skip\");\n"
"                }\n"
"            }\n"
"        }\n"
"    }\n"
"});\n"
"\n"
"// Filter Toggle\n"
"var filterCategories = filter.querySelectorAll(\".filter_topics .filter_content,.filter_media .filter_content\");\n"
"for(var i = 0; i < filterCategories.length; ++i)\n"
"{\n"
"    filterCategories[i].addEventListener(\"click\", function(ev) {\n"
"        var selectedCategory = this.classList[1];\n"
"        filterState[selectedCategory].off = !filterState[selectedCategory].off;\n"
"\n"
"        if(filterState[selectedCategory].off)\n"
"        {\n"
"            this.classList.add(\"off\");\n"
"            var testMarkers = document.querySelectorAll(\".marker.\" + selectedCategory + \", .marker.cat_\" + selectedCategory);\n"
"            for(var j = 0; j < testMarkers.length; ++j)\n"
"            {\n"
"                if(filterState[selectedCategory].type == \"topic\")\n"
"                {\n"
"                    testMarkers[j].classList.remove(\"cat_\" + selectedCategory);\n"
"                    testMarkers[j].classList.add(\"off_\" + selectedCategory);\n"
"                    var markerCategories = testMarkers[j].querySelectorAll(\".category.\" + selectedCategory);\n"
"                    for(var k = 0; k < markerCategories.length; ++k)\n"
"                    {\n"
"                        if(markerCategories[k].classList.contains(selectedCategory))\n"
"                        {\n"
"                            markerCategories[k].classList.add(\"off\");\n"
"                        }\n"
"                    }\n"
"                }\n"
"                else\n"
"                {\n"
"                    testMarkers[j].classList.remove(selectedCategory);\n"
"                    testMarkers[j].classList.add(\"off_\" + selectedCategory);\n"
"                }\n"
"\n"
"                Skipping = 1;\n"
"                if(filterMode == \"exclusive\")\n"
"                {\n"
"                    testMarkers[j].classList.add(\"skip\");\n"
"                }\n"
"                else\n"
"                {\n"
"                    var markerClasses = testMarkers[j].classList;\n"
"                    for(var k = 0; k < markerClasses.length; ++k)\n"
"                    {\n"
"                        if(markerClasses[k] in filterState || markerClasses[k].replace(/^cat_/, \"\") in filterState)\n"
"                        {\n"
"                            Skipping = 0;\n"
"                        }\n"
"                    }\n"
"                    if(Skipping)\n"
"                    {\n"
"                        testMarkers[j].classList.add(\"skip\");\n"
"                    }\n"
"                }\n"
"\n"
"            }\n"
"        }\n"
"        else\n"
"        {\n"
"            this.classList.remove(\"off\");\n"
"            var testMarkers = document.querySelectorAll(\".marker.off_\" + selectedCategory);\n"
"            for(var j = 0; j < testMarkers.length; ++j)\n"
"            {\n"
"                if(filterState[selectedCategory].type == \"topic\")\n"
"                {\n"
"                    testMarkers[j].classList.remove(\"off_\" + selectedCategory);\n"
"                    testMarkers[j].classList.add(\"cat_\" + selectedCategory);\n"
"                    var markerCategories = testMarkers[j].querySelectorAll(\".category.\" + selectedCategory);\n"
"                    for(var k = 0; k < markerCategories.length; ++k)\n"
"                    {\n"
"                        if(markerCategories[k].classList.contains(selectedCategory))\n"
"                        {\n"
"                            markerCategories[k].classList.remove(\"off\");\n"
"                        }\n"
"                    }\n"
"                }\n"
"                else\n"
"                {\n"
"                    testMarkers[j].classList.remove(\"off_\" + selectedCategory);\n"
"                    testMarkers[j].classList.add(selectedCategory);\n"
"                }\n"
"\n"
"                Skipping = 0;\n"
"                if(filterMode == \"inclusive\")\n"
"                {\n"
"                    testMarkers[j].classList.remove(\"skip\");\n"
"                }\n"
"                else\n"
"                {\n"
"                    var markerClasses = testMarkers[j].classList;\n"
"                    for(var k = 0; k < markerClasses.length; ++k)\n"
"                    {\n"
"                        if(markerClasses[k].startsWith(\"off_\"))\n"
"                        {\n"
"                            Skipping = 1;\n"
"                        }\n"
"                    }\n"
"                    if(!Skipping)\n"
"                    {\n"
"                        testMarkers[j].classList.remove(\"skip\");\n"
"                    }\n"
"                }\n"
"            }\n"
"        }\n"
"    });\n"
"}\n"
"\n"
"var refSources = document.querySelectorAll(\".refs .ref\");\n"
"for (var i = 0; i < refSources.length; ++i) {\n"
"    refSources[i].addEventListener(\"click\", function(ev) {\n"
"        if (player) {\n"
"            player.pause();\n"
"        }\n"
"    });\n"
"}\n"
"\n"
"function resetFade()\n"
"{\n"
"    filter.classList.remove(\"responsible\");\n"
"    filter.querySelector(\".filter_mode\").classList.remove(\"responsible\");\n"
"    var responsibleCategories = filter.querySelectorAll(\".filter_content.responsible\");\n"
"    for(var i = 0; i < responsibleCategories.length; ++i)\n"
"    {\n"
"        responsibleCategories[i].classList.remove(\"responsible\");\n"
"    }\n"
"}\n"
"\n"
"var sourceMenus = document.querySelectorAll(\".menu\");\n"
"function onRefChanged(ref, element) {\n"
"    if(element.classList.contains(\"skip\"))\n"
"    {\n"
"        if(!filter.classList.contains(\"responsible\"))\n"
"        {\n"
"            filter.classList.add(\"responsible\");\n"
"        }\n"
"\n"
"        for(var selector = 0; selector < element.classList.length; ++selector)\n"
"        {\n"
"            if(element.classList[selector].startsWith(\"off_\"))\n"
"            {\n"
"                if(!filter.querySelector(\".filter_content.\" + element.classList[selector].replace(/^off_/, \"\")).classList.contains(\"responsible\"))\n"
"                {\n"
"                    filter.querySelector(\".filter_content.\" + element.classList[selector].replace(/^off_/, \"\")).classList.add(\"responsible\");\n"
"                }\n"
"            }\n"
"            if((element.classList[selector].startsWith(\"cat_\") || element.classList[selector] in filterState))\n"
"            {\n"
"                if(!filter.querySelector(\".filter_mode\").classList.add(\"responsible\"))\n"
"                {\n"
"                    filter.querySelector(\".filter_mode\").classList.add(\"responsible\");\n"
"                }\n"
"            }\n"
"            setTimeout(resetFade, 8000);\n"
"        }\n"
"        player.jumpToNextMarker();\n"
"        return;\n"
"    }\n"
"\n"
"    for (var MenuIndex = 0; MenuIndex < sourceMenus.length; ++MenuIndex)\n"
"    {\n"
"        var SetMenu = 0;\n"
"        if (ref !== undefined && ref !== null) {\n"
"            var refElements = sourceMenus[MenuIndex].querySelectorAll(\".refs .ref\");\n"
"            var refs = ref.split(\",\");\n"
"\n"
"            for (var i = 0; i < refElements.length; ++i) {\n"
"                if (refs.includes(refElements[i].getAttribute(\"data-id\"))) {\n"
"                    refElements[i].classList.add(\"current\");\n"
"                    SetMenu = 1;\n"
"                } else {\n"
"                    refElements[i].classList.remove(\"current\");\n"
"                }\n"
"            }\n"
"            if(SetMenu) {\n"
"                sourceMenus[MenuIndex].classList.add(\"current\");\n"
"            } else {\n"
"                sourceMenus[MenuIndex].classList.remove(\"current\");\n"
"            }\n"
"\n"
"        } else {\n"
"            sourceMenus[MenuIndex].classList.remove(\"current\");\n"
"            var refs = sourceMenus[MenuIndex].querySelectorAll(\".refs .ref\");\n"
"            for (var i = 0; i < refs.length; ++i) {\n"
"                refs[i].classList.remove(\"current\");\n"
"            }\n"
"        }\n"
"    }\n"
"}\n"
"        </script>\n"
"    </body>\n"
"</html>\n");

#if 0
            CopyStringToBuffer(&Master,
"        <script>\n"
"            var player = new Player(document.querySelector(\".player_container\"), onRefChanged);\n"
"            window.addEventListener(\"resize\", function() { player.updateSize(); });\n"
"                document.addEventListener(\"keypress\", function(ev) {\n"
"                    switch (ev.key) {\n"
"                        case 'n':\n"
"                        case 'd':\n"
"                        case 's': {\n"
"                                player.jumpToNextMarker();\n"
"                        } break;\n"
"\n"
"                        case 'p':\n"
"                        case 'a':\n"
"                        case 'w': {\n"
"                                player.jumpToPrevMarker();\n"
"                        } break;\n"
"    }\n"
"});\n"
"\n"
"var refTimecodes = document.querySelectorAll(\".refs .ref .timecode\");\n"
"for (var i = 0; i < refTimecodes.length; ++i) {\n"
"    refTimecodes[i].addEventListener(\"click\", function(ev) {\n"
"        if (player) {\n"
"            var time = ev.currentTarget.getAttribute(\"data-timestamp\");\n"
"            player.setTime(parseInt(time, 10));\n"
"            player.play();\n"
"            ev.preventDefault();\n"
"            ev.stopPropagation();\n"
"            return false;\n"
"        }\n"
"    });\n"
"}\n"
"\n"
"var refSources = document.querySelectorAll(\".refs .ref\");\n"
"for (var i = 0; i < refSources.length; ++i) {\n"
"    refSources[i].addEventListener(\"click\", function(ev) {\n"
"        if (player) {\n"
"            player.pause();\n"
"        }\n"
"    });\n"
"}\n"
"\n"
"function onRefChanged(ref) {\n"
"    var sourceMenus = document.querySelectorAll(\".menu\");\n"
"    for (var MenuIndex = 0; MenuIndex < sourceMenus.length; ++MenuIndex)\n"
"        {\n"
"        var SetMenu = 0;\n"
"        if (ref !== undefined && ref !== null) {\n"
"            var refElements = sourceMenus[MenuIndex].querySelectorAll(\".refs .ref\");\n"
"            var refs = ref.split(\",\");\n"
"\n"
"            for (var i = 0; i < refElements.length; ++i) {\n"
"                if (refs.includes(refElements[i].getAttribute(\"data-id\"))) {\n"
"                    refElements[i].classList.add(\"current\");\n"
"                    SetMenu = 1;\n"
"                } else {\n"
"                    refElements[i].classList.remove(\"current\");\n"
"                }\n"
"            }\n"
"            if(SetMenu) {\n"
"                sourceMenus[MenuIndex].classList.add(\"current\");\n"
"            } else {\n"
"                sourceMenus[MenuIndex].classList.remove(\"current\");\n"
"            }\n"
"\n"
"        } else {\n"
"            sourceMenus[MenuIndex].classList.remove(\"current\");\n"
"            var refs = sourceMenus[MenuIndex].querySelectorAll(\".refs .ref\");\n"
"            for (var i = 0; i < refs.length; ++i) {\n"
"                refs[i].classList.remove(\"current\");\n"
"            }\n"
"        }\n"
"    }\n"
"}\n"
"        </script>\n"
"    </body>\n"
"</html>\n");
#endif

            FILE *OutFile;
            if(!(OutFile = fopen("out.html", "w")))
            {
                perror(Args[0]);
                hmml_free(&HMML);
                free(MemoryArena);
                return 1;
            }
            fwrite(Master.Location, Master.Ptr - Master.Location, 1, OutFile);
            fclose(OutFile);

           DeclaimBuffer(&Master, &ClaimedMemory);
        }
        else
        {
            fprintf(stderr, "%s:%d: %s\n", Args[FileIndex], HMML.error.line, HMML.error.message);
        }
        hmml_free(&HMML);
    }
    free(MemoryArena);
}
